#include <QDebug>

#include "daqthread.h"

#define DAQmxErrChk(functionCall) if( DAQmxFailed(error=(functionCall)) ) goto Error; else

DAQThread::DAQThread(QObject *parent) :
    QThread(parent)
{
    stopped = false;

    state = State_IDLE;
    EvtQue = new QQueue<int>();

    // cannot Init (create) DAQmxTask, when Constructer
}

DAQThread::~DAQThread()
{
    delete EvtQue;
}

void DAQThread::stop()
{
    stopped = true;
}

void DAQThread::startPulseOutput(QString &devname, float freq, float duty, int continue_mode = 1, unsigned long long int pulsecount = 1)
{
    QString tmpname = devname + "/ctr0";
    // qDebug() << "counter name  is " << tmpname;
    ctrname = tmpname.toLocal8Bit();
    pulse_freq = freq;
    pulse_duty = (float64)(duty / 100.0);
    cont_mode = continue_mode ? DAQmx_Val_ContSamps : DAQmx_Val_FiniteSamps;
    pulse_count = pulsecount;
    EvtQue->enqueue(Ev_Start);
}

void DAQThread::stopPulseOutput()
{
    EvtQue->enqueue(Ev_Stop);
}

void DAQThread::changePulseOutput(float freq, float duty)
{
    pulse_freq = freq;
    pulse_duty = (float64)(duty / 100.0);
    EvtQue->enqueue(Ev_ChgPulse);
}

void DAQThread::run()
{
    int32       error = 0;
    char        errBuff[2048] = { '\0' };
//    char        ctrname[128] = "Dev1/ctr0";

    taskHandle = 0;
    ctrname = "Dev1/ctr0";
    pulse_freq = (float64)1000.0;
    pulse_duty = (float64)0.50; // 0.50 is 50%
    int event = 0;
    state = State_IDLE;

    uInt32 serialnumber;
    DAQmxBaseGetDevSerialNum("Dev1", &serialnumber);
    qDebug() << QString("%1 SerialNumber is 0x%2").arg("Dev1").arg(serialnumber, 8, 16, QChar('0'));

//    DAQmxBaseResetDevice("Dev1");

    // signal - daqmxbase ready
    emit daqmxbase_ready();

    while(!stopped) {
        if (EvtQue->isEmpty()) {
            if (state == State_OUTPUT) {
                bool32 done = 0;
                DAQmxBaseIsTaskDone(taskHandle, &done);
                if (done) {
                    DAQmxErrChk (DAQmxBaseClearTask(taskHandle));
                    taskHandle = 0;
                    emit pulse_output_stopped();
                    qDebug() << "done, stop pulse output."; // stop Pulse
                    state = State_IDLE;
                }
            }
            msleep(100);
            continue;  // Jump to Next Loop
        }

        event = EvtQue->dequeue();
        switch (state) {
        case State_IDLE:
            if (event == Ev_Start) {
                qDebug() << QString("start pulse output: freq = %1, duty = %2").arg(pulse_freq).arg(pulse_duty);
                //    int32 DAQmxBaseCreateCOPulseChanFreq (TaskHandle taskHandle, const char counter[ ], const char nameToAssignToChannel[ ], int32 units, int32 idleState, float64 initialDelay, float64 freq, float64 dutyCycle);
                DAQmxErrChk (DAQmxBaseCreateTask("", &taskHandle));
                DAQmxErrChk (DAQmxBaseCreateCOPulseChanFreq (taskHandle, ctrname, "", DAQmx_Val_Hz, DAQmx_Val_Low, (float64)0.0, pulse_freq, pulse_duty));
                //    int32 DAQmxBaseCfgImplicitTiming (TaskHandle taskHandle, int32 sampleMode, uInt64 sampsPerChanToAcquire);
                DAQmxErrChk (DAQmxBaseCfgImplicitTiming (taskHandle, cont_mode, pulse_count));
                //DAQmxErrChk (DAQmxBaseExportSignal(taskHandle, DAQmx_Val_CounterOutputEvent, "/Dev1/PFI6")); // not worked! , default "/Dev1/PFI4"
                DAQmxErrChk (DAQmxBaseStartTask (taskHandle));
                emit pulse_output_started();
                state = State_OUTPUT;
            }
            break;
        case State_OUTPUT:
            if (event == Ev_Stop) {
                DAQmxErrChk (DAQmxBaseStopTask (taskHandle));
                msleep(200);
                DAQmxErrChk (DAQmxBaseClearTask (taskHandle));
                taskHandle = 0;
                emit pulse_output_stopped();
                qDebug() << "stop pulse output."; // stop Pulse
                state = State_IDLE;
            } else if (event == Ev_ChgPulse) {
                qDebug() << QString("change pulse freq, duty : freq = %1, ducy = %2").arg(pulse_freq).arg(pulse_duty);
                DAQmxErrChk (DAQmxBaseStopTask (taskHandle));
                msleep(200);
                DAQmxErrChk (DAQmxBaseClearTask (taskHandle));
                taskHandle = 0;
                // msleep(200);
                DAQmxErrChk (DAQmxBaseCreateTask("", &taskHandle));
                //    int32 DAQmxBaseCreateCOPulseChanFreq (TaskHandle taskHandle, const char counter[ ], const char nameToAssignToChannel[ ], int32 units, int32 idleState, float64 initialDelay, float64 freq, float64 dutyCycle);
                DAQmxErrChk (DAQmxBaseCreateCOPulseChanFreq (taskHandle, ctrname, "", DAQmx_Val_Hz, DAQmx_Val_Low, (float64)0.0, pulse_freq, pulse_duty));
                //    int32 DAQmxBaseCfgImplicitTiming (TaskHandle taskHandle, int32 sampleMode, uInt64 sampsPerChanToAcquire);
                DAQmxErrChk (DAQmxBaseCfgImplicitTiming (taskHandle, cont_mode, pulse_count));
                DAQmxErrChk (DAQmxBaseStartTask (taskHandle));
            }
            break;
        default:
            break;
        }
        msleep(100);
    }

Error:
    if( DAQmxFailed(error) ) {
        DAQmxBaseGetExtendedErrorInfo(errBuff,2048);
        qDebug() << errBuff;
        emit daqmxbase_error(QString(errBuff));
    }
    if( taskHandle!=0 )  {
        /*********************************************/
        // DAQmx Stop Code
        /*********************************************/
        DAQmxBaseStopTask(taskHandle);
        msleep(200);
        DAQmxBaseClearTask(taskHandle);
        emit daqmxbase_final();
    }
}
